import 'dart:convert';
import 'dart:developer';
import 'package:dio/dio.dart';
import 'package:flutter/material.dart';

class NetworkOption extends BaseOptions {
  ///called on start every request
  Function onStartDefault = () {};

  ///called on End every request
  Function onEndDefault = () {};

  ///called on Success every request
  void Function(NetworkResponse) onSuccessDefault = (NetworkResponse response) {};

  ///called on Failed every request
  void Function(NetworkResponse) onFailedDefault = (NetworkResponse response) {};

  CancelToken globalCancelToken;

  ///String token which is used in header
  String token;

  ///called on token expire
  bool Function(NetworkResponse) tokenExpireRule = (res) =>res.responseCode==403;
  void Function(NetworkResponse) onTokenExpire = (_) {print("Token Expired");};

  /// a function to extract error as string
  String Function(dynamic error) errorMsgExtractor = (error) => error.toString();
  String Function(dynamic msg) successMsgExtractor = (msg) => msg.toString();

  static final NetworkOption _singleton = NetworkOption._internal();

  NetworkOption._internal();

  factory NetworkOption() {
    return _singleton;
  }

  static setAccessToken(String token) {
    log("Token is set: $token");
    _singleton.headers.update("Authorization",(v)=>"Bearer $token",ifAbsent:
        ()=>_singleton.headers.putIfAbsent("Authorization", ()=>"Bearer $token") );
    _singleton.token = token;
  }

  static initialize({
    String baseUrl,
    int timeout = 15000,
    String token,
    Function onStartDefault,
    Function onEndDefault,
    Function onSuccessDefault,
    Function onFailedDefault,
    CancelToken globalCancelToken,
    Map<String, dynamic> headers,
    bool Function(NetworkResponse) tokenExpireRule,
    void Function(NetworkResponse) onTokenExpire,
    String Function(dynamic error) errorMsgExtractor,
    String Function(dynamic msg) successMsgExtractor,
  }) {
    if (onStartDefault != null) {
      _singleton.onStartDefault = onStartDefault;
    }
    if (onEndDefault != null) {
      _singleton.onEndDefault = onEndDefault;
    }
    if (onSuccessDefault != null) {
      _singleton.onSuccessDefault = onSuccessDefault;
    }
    if (onFailedDefault != null) {
      _singleton.onFailedDefault = onFailedDefault;
    }
    if (onTokenExpire != null) {
      _singleton.onTokenExpire = onTokenExpire;
    }
    if (tokenExpireRule != null) {
      _singleton.tokenExpireRule = tokenExpireRule;
    }
    if (baseUrl != null) {
      _singleton.baseUrl = baseUrl;
    }
    if (headers != null) {
      _singleton.headers = headers;
    }
    if (timeout != null) {
      _singleton.connectTimeout = timeout;
      _singleton.sendTimeout = timeout;
      _singleton.receiveTimeout = timeout;
    }

    if (globalCancelToken != null) {
      _singleton.globalCancelToken = globalCancelToken;
    }

    _singleton.token = token;

    ///You can set LogInterceptor to print request/response log automatically
//    _singleton.dio.interceptors.add(LogInterceptor(responseBody: false));
    if (errorMsgExtractor != null) {
      _singleton.errorMsgExtractor = errorMsgExtractor;
    }
    if (successMsgExtractor != null) {
      _singleton.successMsgExtractor = errorMsgExtractor;
    }
  }
}

class NetworkResponse {
  final bool responseStatus;
  final int responseCode;
  final String responseTitle;
  Future<NetworkResponse> Function() retryFunction;
  dynamic responseBody;
  dynamic responseDetails;
  String extractedMessage;

  NetworkResponse(
      {this.responseStatus,
        this.responseCode,
        this.responseTitle,
        this.responseBody,
        this.retryFunction,
        this.responseDetails,
        this.extractedMessage});

  @override
  String toString() {
    return "$responseCode $responseTitle : $extractedMessage";
  }
}

class NetworkRequest {
  final String api;
  final dynamic data;
  final Dio dio;
  final Duration timeOut;
  final NetworkOption options;
  final bool callDefaults;

  NetworkRequest._({
    @required this.api,
    this.data,
    this.timeOut = const Duration(seconds: 30),
    this.options,
    this.callDefaults = true,
  }) : dio = Dio(options);

  factory NetworkRequest({@required api, data,bool callDefaults=true, timeOut = const Duration(seconds: 30)}) {
    NetworkOption option = NetworkOption();
    return NetworkRequest._(api: api, data: data, timeOut: timeOut, options: option,callDefaults: callDefaults);
  }

  Future<NetworkResponse> post() async {
    if(callDefaults){
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      log("Posting ${jsonEncode(data)} in ${jsonEncode(api)}");
      Response res = await dio.post(api, data: data, cancelToken: options.globalCancelToken).catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.CONNECT_TIMEOUT;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode > 199 && res.statusCode < 300,
          responseCode: res.statusCode,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response.statusCode > 199 && e.response.statusCode < 300,
            responseCode: e.response.statusCode,
            responseTitle: e.response.statusMessage,
            responseBody: e.response.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response.data));
      } else if(e is DioError ){
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Error",
            responseBody: "Something went wrong",
            responseDetails: e,
            extractedMessage: "Something went wrong");
      } else{
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -1,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }

    response.retryFunction = post;
    log(response.toString());

    if(callDefaults){
      if(response.responseStatus){
        options.onSuccessDefault(response);
      }else{
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    }
    if(options.tokenExpireRule(response)){
      options.onTokenExpire(response);
    }


    return response;
  }
}
